// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"os"
	"os/signal"
	"path"
	"syscall"

	gnatsd "github.com/nats-io/gnatsd/server"
	"github.com/slicebit/qb"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/xbus/server"
)

var serveCmd = &cobra.Command{
	Use:   "serve",
	Short: "Run Xbus servers",
	Long: `Run one or several of the Xbus servers.
The servers are:
- nats: A genuine nats server
- broker: Exchanges envelopes with actors
- control: Handle control API`,
	PersistentPreRun: preRunCheckCertCertDir,
	Run:              xbusServeRun,
}

func init() {
	RootCmd.AddCommand(serveCmd)
	flags := serveCmd.PersistentFlags()

	flags.Bool(
		"debug-sql-queries", false,
		"Logs sql queries to debug. Very verbose",
	)

	flags.Bool(
		"debug-sql-bindings", false,
		"Logs sql bindings to debug. Very verbose",
	)

	flags.Bool(
		"nats", true,
		"Start a NATS server",
	)

	flags.Bool(
		"broker", true,
		"Start the broker service",
	)

	flags.Bool(
		"clientapi", true,
		"Start the clientapi service",
	)

	flags.Bool(
		"emitter", true,
		"Start the emitter service",
	)

	flags.Bool(
		"control", true,
		"Start the control service",
	)

	flags.Bool(
		"director", true,
		"Start the director service",
	)

	flags.Bool(
		"envelopestorage", true,
		"Start the envelopestorage service",
	)

	flags.Bool(
		"postmortem", true,
		"Start the postmortem service",
	)

	flags.Bool(
		"processstate", true,
		"Start the processstate service",
	)

	flags.Bool(
		"registration", true,
		"Start the registration service",
	)

	flags.Bool(
		"sentinel", true,
		"Start the sentinel service",
	)

	flags.String(
		"gnatsd.http_host", "",
		"User host for http monitoring",
	)
	flags.Int(
		"gnatsd.http_port", 0,
		"User port for http monitoring",
	)
	flags.Int(
		"gnatsd.https_port", 0,
		"User port for https monitoring",
	)

	flags.String(
		"sentry.dsn", "",
		"DSN of a sentry instance",
	)

	flags.String(
		"sentry.environment", "",
		"DSN of a sentry instance",
	)

	flags.String(
		"metrics-http-host", "",
		"User host for http metrics (prometheus)",
	)

	flags.Int(
		"metrics-http-port", 0,
		"User port for http metrics (prometheus)",
	)

	viper.SetDefault("debug-sql-queries", false)
	viper.BindPFlag("debug-sql-queries", flags.Lookup("debug-sql-queries"))

	viper.SetDefault("debug-sql-bindings", false)
	viper.BindPFlag("debug-sql-bindings", flags.Lookup("debug-sql-bindings"))

	viper.SetDefault("serve.nats", true)
	viper.BindPFlag("serve.nats", flags.Lookup("nats"))

	viper.SetDefault("serve.broker", true)
	viper.BindPFlag("serve.broker", flags.Lookup("broker"))

	viper.SetDefault("serve.clientapi", true)
	viper.BindPFlag("serve.clientapi", flags.Lookup("clientapi"))

	viper.SetDefault("serve.control", true)
	viper.BindPFlag("serve.control", flags.Lookup("control"))

	viper.SetDefault("serve.director", true)
	viper.BindPFlag("serve.director", flags.Lookup("director"))

	viper.SetDefault("serve.envelopestorage", true)
	viper.BindPFlag("serve.envelopestorage", flags.Lookup("envelopestorage"))

	viper.SetDefault("serve.postmortem", true)
	viper.BindPFlag("serve.postmortem", flags.Lookup("postmortem"))

	viper.SetDefault("serve.processstate", true)
	viper.BindPFlag("serve.processstate", flags.Lookup("processstate"))

	viper.SetDefault("serve.registration", true)
	viper.BindPFlag("serve.registration", flags.Lookup("registration"))

	viper.SetDefault("serve.sentinel", true)
	viper.BindPFlag("serve.sentinel", flags.Lookup("sentinel"))

	viper.BindPFlag("gnatsd.http_host", flags.Lookup("gnatsd.http_host"))
	viper.BindPFlag("gnatsd.http_port", flags.Lookup("gnatsd.http_port"))
	viper.BindPFlag("gnatsd.https_port", flags.Lookup("gnatsd.https_port"))

	viper.BindPFlag("sentry.dsn", flags.Lookup("sentry.dsn"))
	viper.BindPFlag("sentry.environment", flags.Lookup("sentry.environment"))
	viper.BindPFlag("metrics.http-host", flags.Lookup("metrics-http-host"))
	viper.BindPFlag("metrics.http-port", flags.Lookup("metrics-http-port"))
}

func xbusServeRun(cmd *cobra.Command, args []string) {
	certDir := viper.GetString("cert-dir")

	logFlags := qb.LDefault
	if viper.GetBool("debug-sql-queries") {
		logFlags |= qb.LQuery
	}
	if viper.GetBool("debug-sql-bindings") {
		logFlags |= qb.LBindings
	}

	opts := server.Options{
		DatabaseDSN:                    viper.GetString("database.dsn"),
		DatabaseMaxConnections:         viper.GetInt("database.max-connections"),
		ProcessStorageDBMaxConnections: viper.GetInt("storage.process.db.max-connections"),
		DatabaseLogFlags:               logFlags,
		RootCA:                         path.Join(certDir, "ServerCA.crt"),
		ServerCert:                     path.Join(certDir, "ServerCA.crt"),
		ServerKey:                      path.Join(certDir, "ServerCA.key"),
		ClientCA:                       path.Join(certDir, "ClientCA.crt"),
		ClientCAKey:                    path.Join(certDir, "ClientCA.key"),
		TLSTimeout:                     viper.GetFloat64("tls.timeout"),
		MetricsHTTPHost:                viper.GetString("metrics.http-host"),
		MetricsHTTPPort:                viper.GetInt("metrics.http-port"),
		NATSHost:                       viper.GetString("nats.host"),
		NATSPort:                       viper.GetInt("nats.port"),
		NATSDebug:                      viper.GetBool("debug"),
		NATSTrace:                      viper.GetBool("trace"),
		SentryDSN:                      viper.GetString("sentry.dsn"),
		SentryEnvironment:              viper.GetString("sentry.environment"),
		ServeNATS:                      viper.GetBool("serve.nats"),
		ServeBroker:                    viper.GetBool("serve.broker"),
		ServeClientAPI:                 viper.GetBool("serve.clientapi"),
		ServeControl:                   viper.GetBool("serve.control"),
		ServeDirector:                  viper.GetBool("serve.director"),
		ServeEnvelopeStorage:           viper.GetBool("serve.envelopestorage"),
		ServePostmortem:                viper.GetBool("serve.postmortem"),
		ServeProcessState:              viper.GetBool("serve.processstate"),
		ServeRegistration:              viper.GetBool("serve.registration"),
		ServeSentinel:                  viper.GetBool("serve.sentinel"),
		GnatsdOptions: gnatsd.Options{
			HTTPHost:  viper.GetString("gnatsd.http_host"),
			HTTPPort:  viper.GetInt("gnatsd.http_port"),
			HTTPSPort: viper.GetInt("gnatsd.https_port"),
		},
	}

	s, err := server.NewServer(opts)
	if err != nil {
		Logger.Errorf("Could not setup the server: %s", err)
		os.Exit(1)
	}
	s.SetLogger(Logger)

	if err := s.Startup(); err != nil {
		Logger.Errorf("Could not start the server: %s", err)
		os.Exit(1)
	}

	Logger.Noticef("Server started, now waiting for a signal")

	// TODO do something a little more elaborate so that interrupt during startup
	// is gracefully handlded
	signals := make(chan os.Signal, 1)
	signal.Notify(signals, os.Interrupt, syscall.SIGTERM)

	sig := <-signals

	Logger.Noticef("Signal '%s' caught, shutting down", sig)

	s.Shutdown()
}
