// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"context"
	"errors"
	"fmt"
	"strings"
	"time"

	"github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/go-xbus/tools"
)

var (
	// ErrNotFound Not match found
	ErrNotFound = errors.New("Not found")
	// ErrInvalidGraph The graph is not valid
	ErrInvalidGraph = errors.New("Invalid graph")
)

// ValidationMessagesList A list of validation messages that implements
// the 'error' interface
type ValidationMessagesList []string

// Error return the combined list of messages in one string
func (l ValidationMessagesList) Error() string {
	return "Graph Validation error: \n* " + strings.Join(l, "\n* ")
}

func wrapAPIError(err *nrpc.Error) error {
	if err == nil {
		return nil
	}
	return fmt.Errorf("%s: %s", err.Type, errors.New(err.Message))
}

// ActorCreate create actors
func (c *Client) ActorCreate(actors []api.Actor) ([]api.Actor, error) {
	r, err := c.Control.Actor.Create(control.ActorRequest{ActorList: actors})
	return r.GetActorList(), err
}

// ActorDelete deletes actors
func (c *Client) ActorDelete(ids []api.UUID) error {
	var actors []api.Actor
	for _, id := range ids {
		actors = append(actors, api.Actor{ID: id})
	}
	_, err := c.Control.Actor.Delete(control.ActorRequest{ActorList: actors})
	return err
}

// ActorList lists actors.
func (c *Client) ActorList(actors []api.Actor) ([]api.Actor, error) {
	r, err := c.Control.Actor.Query(control.ActorRequest{ActorList: actors})
	return r.GetActorList(), err
}

// ActorFind finds an actor
func (c *Client) ActorFind(identifier string) (api.Actor, error) {
	var filter api.Actor
	if id, err := api.UUIDFromString(identifier); err == nil {
		filter.ID = id
	} else {
		filter.Name = identifier
	}
	list, err := c.ActorList([]api.Actor{filter})
	if err != nil {
		return api.Actor{}, err
	}

	if len(list) == 0 {
		return api.Actor{}, ErrNotFound
	}
	if len(list) > 1 {
		panic("received multiple actors for " + identifier)
	}
	return list[0], nil
}

// ActorAccept accepts a pending actor
func (c *Client) ActorAccept(id api.UUID) (api.Actor, error) {
	r, err := c.Control.Actor.Update(
		control.ActorRequest{
			ActorList: []api.Actor{{
				ID:     id,
				Status: api.Actor_ENABLED,
			}},
		},
	)
	if err != nil {
		return api.Actor{}, err
	}
	return r.GetActorList()[0], nil
}

// ActorReject rejects a pending actor
func (c *Client) ActorReject(id api.UUID) (api.Actor, error) {
	r, err := c.Control.Actor.Update(
		control.ActorRequest{
			ActorList: []api.Actor{{
				ID:     id,
				Status: api.Actor_REJECTED,
			}},
		},
	)
	if err != nil {
		return api.Actor{}, err
	}
	return r.GetActorList()[0], nil
}

// EnvelopePurge removes unreferenced envelopes
func (c *Client) EnvelopePurge(ctx context.Context, cb func(count uint32)) error {
	return c.Control.Envelope.Purge(
		ctx,
		func(ctx context.Context, p control.TaskProgress) {
			if cb != nil {
				cb(p.Progression)
			}
		})
}

// PipelineQuery searches for pipelines
func (c *Client) PipelineQuery(name string) ([]api.PipelineInfo, error) {
	rep, err := c.Control.Pipeline.Query(control.PipelineQueryRequest{Name: name})
	return rep.GetPipelineList(), err
}

// PipelineSave saves a draft pipeline to the server
func (c *Client) PipelineSave(
	info api.PipelineInfo, graph string,
) (api.PipelineInfo, string, []string, error) {
	rep, err := c.Control.Pipeline.Save(control.PipelineSaveRequest{
		Info:  info,
		Graph: graph,
	})
	if err == nil && !rep.IsValid {
		err = ErrInvalidGraph
	}
	return rep.GetInfo(), rep.GetGraph(), rep.GetValidationMessages(), err
}

// PipelineSetStatus changes the status of a pipeline
func (c *Client) PipelineSetStatus(info api.PipelineInfo) (api.PipelineInfo, error) {
	rep, err := c.Control.Pipeline.SetStatus(info)
	if err == nil && !rep.IsValid {
		err = ValidationMessagesList(rep.GetValidationMessages())
	}
	return rep.GetInfo(), err
}

// PipelineGet returns a pipeline definition
func (c *Client) PipelineGet(info api.PipelineInfo) (string, error) {
	rep, err := c.Control.Pipeline.GetGraph(info)
	return rep.GetValue(), err
}

// ProcessQuery queries processes
func (c *Client) ProcessQuery(
	ctx context.Context, filter api.ProcessFilter,
	cb func([]api.Process),
) error {
	err := c.Control.Process.Query(
		ctx,
		control.ProcessQueryRequest{Filter: filter},
		func(ctx context.Context, r control.ProcessQueryReply) {
			cb(r.GetProcessList())
		},
	)
	return err
}

// ProcessControl pause/resume/cancel a process
func (c *Client) ProcessControl(processID api.UUID, command control.ProcessControlRequest_Command) error {
	return c.Control.Process.Control(
		control.ProcessControlRequest{
			ProcessID: processID,
			Command:   command,
		},
	)
}

// ProcessExport exports processes data (definition, state, logs, envelopes...)
func (c *Client) ProcessExport(processIDs []api.UUID) ([]string, error) {
	rep, err := c.Control.Process.Export(control.ProcessExportRequest{
		ProcessIDs: processIDs,
	})
	return rep.GetProcessList(), err
}

// ProcessPurge purges processes data (definition, state, routes, logs)
func (c *Client) ProcessPurge(ctx context.Context, processIDs []api.UUID, cb func(progression uint32, maxProgression uint32)) error {
	return c.Control.Process.Purge(
		ctx,
		control.ProcessPurgeRequest{ProcessIDs: processIDs},
		func(ctx context.Context, p control.TaskProgress) {
			if cb != nil {
				cb(p.Progression, p.MaxProgression)
			}
		},
	)
}

// AccountList lists the bus accounts
func (c *Client) AccountList() ([]api.Account, error) {
	rep, err := c.Control.Account.List(control.AccountListRequest{})
	return rep.GetAccountList(), err
}

// AccountFind finds an account either by name or fingerprint
func (c *Client) AccountFind(identifier string) (api.Account, error) {
	accounts, err := c.AccountList()
	if err != nil {
		return api.Account{}, err
	}

	for _, account := range accounts {
		if tools.GetCSRHash(account.CSR, true) == identifier {
			return account, nil
		}
		if account.Name == identifier {
			return account, nil
		}
	}

	return api.Account{}, ErrNotFound
}

// AccountAccept accepts a pending account
func (c *Client) AccountAccept(id api.UUID, expire time.Time) (api.Account, error) {
	return c.Control.Account.Update(
		control.AccountUpdateRequest{
			Account: api.Account{
				ID:     id,
				Status: api.Account_ENABLED,
			},
			Expire: expire,
		},
	)
}

// AccountReject rejects a pending account
func (c *Client) AccountReject(id api.UUID) (api.Account, error) {
	return c.Control.Account.Update(
		control.AccountUpdateRequest{
			Account: api.Account{
				ID:     id,
				Status: api.Account_DISABLED,
			},
		},
	)
}

// AccountDelete deletes accounts
func (c *Client) AccountDelete(id api.UUID) error {
	return c.Control.Account.Delete(control.AccountIDRequest{AccountID: id})
}

// AccountRenewAPIKey deletes accounts
func (c *Client) AccountRenewAPIKey(id api.UUID) (api.Account, error) {
	return c.Control.Account.RenewAPIKey(control.AccountIDRequest{AccountID: id})
}

// PMProcessQuery returns a list of ended processes
func (c *Client) PMProcessQuery(ctx context.Context, level api.LogLevel, includeClosed bool, processIDs []api.UUID, returnLogs bool) ([]api.PMProcess, error) {
	var list []api.PMProcess
	err := c.Control.PMProcess.Query(ctx,
		control.PMProcessQueryRequest{
			Level:         level,
			IncludeClosed: includeClosed,
			ProcessIDs:    processIDs,
			ReturnLogs:    returnLogs,
		}, func(ctx context.Context, rep control.PMProcessQueryReply) {
			list = append(list, rep.GetPMProcessList()...)
		},
	)
	return list, err
}

// PMProcessSetStatus changes the postmortem status of a process
func (c *Client) PMProcessSetStatus(processID api.UUID, status api.PMProcess_Status, comment string) error {
	return c.Control.PMProcess.SetStatus(control.PMProcessSetStatusRequest{
		ProcessID: processID,
		Status:    status,
		Comment:   comment,
	})
}

// LogsPurge purges logs
func (c *Client) LogsPurge(ctx context.Context, before time.Time) (count int64, err error) {
	err = c.Control.Logs.Purge(
		ctx,
		control.LogsPurgeRequest{Before: before},
		func(ctx context.Context, reply control.LogsPurgeReply) {
			count = reply.GetCount()
		})
	return
}
